;; Generated package description from filechooser.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "filechooser" "0.2.4" "An xdg-desktop-portal filechooser" '((emacs "28.1") (compat "29.1")) :commit "e55c5ea294bbb2c67ab5c51d9489922a83c22456" :authors '(("rahguzar" . "rahguzar@mailbox.org")) :maintainer '("rahguzar" . "rahguzar@mailbox.org") :keywords '("convenience" "files" "tools" "unix") :url "https://codeberg.org/rahguzar/filechooser")
